using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Text;

using log4net;

namespace ServicesCtl
{
    class ServiceControlManager
    {
        [DllImport("advapi32.dll", EntryPoint = "OpenSCManagerW", ExactSpelling = true, CharSet = CharSet.Unicode, SetLastError = true)]
        public static extern IntPtr OpenSCManager(string machineName, string databaseName, uint dwAccess);

        [DllImport("kernel32.dll")]
        private unsafe static extern int FormatMessage(int dwFlags, int lpSource,
            int dwMessageId, int dwLanguageId, ref String lpBuffer, int nSize, int Arguments);

        [DllImport("advapi32.dll", EntryPoint = "EnumServicesStatusW", ExactSpelling = true, SetLastError = true)]
        static extern bool EnumServicesStatus(IntPtr hSCManager,
                            SERVICE_TYPES dwServiceType,
                            SERVICE_STATES dwServiceState,
                            IntPtr lpServices,
                            int cbBufSize,
                            ref int pcbBytesNeeded,
                            ref int lpServicesReturned,
                            ref int lpResumeHandle);

        [DllImport("advapi32", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        public static extern bool StartService(
                        IntPtr hService,
                    int dwNumServiceArgs,
                    string[] lpServiceArgVectors
                    );

        [DllImport("advapi32.dll", SetLastError = true, CharSet = CharSet.Auto)]
        static extern IntPtr OpenService(IntPtr hSCManager, string lpServiceName, uint dwDesiredAccess);

        [DllImport("advapi32.dll", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        public static extern bool ControlService(IntPtr hService,
                               SERVICE_CONTROL dwControl,
                               ref SERVICE_STATUS lpServiceStatus
                                                            );

        [DllImport("advapi32.dll", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        public static extern bool CloseServiceHandle(IntPtr hSCObject);

        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        public struct ENUM_SERVICE_STATUS
        {
            public static readonly int SizeOf = Marshal.SizeOf(typeof(ENUM_SERVICE_STATUS));
            [MarshalAs(System.Runtime.InteropServices.UnmanagedType.LPWStr)]
            public string pServiceName;
            [MarshalAs(System.Runtime.InteropServices.UnmanagedType.LPWStr)]
            public string pDisplayName;
            public SERVICE_STATUS ServiceStatus;
        }
        [StructLayout(LayoutKind.Sequential, Pack = 1)]
        public struct SERVICE_STATUS
        {
            public static readonly int SizeOf = Marshal.SizeOf(typeof(SERVICE_STATUS));
            public SERVICE_TYPES dwServiceType;
            public SERVICE_STATE dwCurrentState;
            public uint dwControlsAccepted;
            public uint dwWin32ExitCode;
            public uint dwServiceSpecificExitCode;
            public uint dwCheckPoint;
            public uint dwWaitHint;
        }
        [Flags]
        public enum SERVICE_STATE : int
        {
            SERVICE_ACTIVE = 0x00000001,
            SERVICE_INACTIVE = 0x00000002,
            SERVICE_STATE_ALL = SERVICE_ACTIVE | SERVICE_INACTIVE
        }
        public enum SERVICE_CONTROL : uint
        {
            STOP = 0x00000001,
            PAUSE = 0x00000002,
            CONTINUE = 0x00000003,
            INTERROGATE = 0x00000004,
            SHUTDOWN = 0x00000005,
            PARAMCHANGE = 0x00000006,
            NETBINDADD = 0x00000007,
            NETBINDREMOVE = 0x00000008,
            NETBINDENABLE = 0x00000009,
            NETBINDDISABLE = 0x0000000A,
            DEVICEEVENT = 0x0000000B,
            HARDWAREPROFILECHANGE = 0x0000000C,
            POWEREVENT = 0x0000000D,
            SESSIONCHANGE = 0x0000000E
        }
        public enum SERVICE_STATES : int
        {
            SERVICE_ACTIVE = 0x00000001,
            SERVICE_INACTIVE = 0x00000002,
            SERVICE_STATE_ALL = SERVICE_ACTIVE | SERVICE_INACTIVE
        }
        public enum SERVICE_TYPES : int
        {
            SERVICE_KERNEL_DRIVER = 0x00000001,
            SERVICE_FILE_SYSTEM_DRIVER = 0x00000002,
            SERVICE_ADAPTER = 0x00000004,
            SERVICE_RECOGNIZER_DRIVER = 0x00000008,
            SERVICE_DRIVER = SERVICE_KERNEL_DRIVER |
                SERVICE_FILE_SYSTEM_DRIVER |
                SERVICE_RECOGNIZER_DRIVER,

            SERVICE_WIN32_OWN_PROCESS = 0x00000010,
            SERVICE_WIN32_SHARE_PROCESS = 0x00000020,
            SERVICE_WIN32 = SERVICE_WIN32_OWN_PROCESS |
                SERVICE_WIN32_SHARE_PROCESS,
        }
        public enum ACCESS_MASK : uint
        {
            DELETE = 0x00010000,
            READ_CONTROL = 0x00020000,
            WRITE_DAC = 0x00040000,
            WRITE_OWNER = 0x00080000,
            SYNCHRONIZE = 0x00100000,

            STANDARD_RIGHTS_REQUIRED = 0x000f0000,

            STANDARD_RIGHTS_READ = 0x00020000,
            STANDARD_RIGHTS_WRITE = 0x00020000,
            STANDARD_RIGHTS_EXECUTE = 0x00020000,

            STANDARD_RIGHTS_ALL = 0x001f0000,

            SPECIFIC_RIGHTS_ALL = 0x0000ffff,

            ACCESS_SYSTEM_SECURITY = 0x01000000,

            MAXIMUM_ALLOWED = 0x02000000,

            GENERIC_READ = 0x80000000,
            GENERIC_WRITE = 0x40000000,
            GENERIC_EXECUTE = 0x20000000,
            GENERIC_ALL = 0x10000000,

            DESKTOP_READOBJECTS = 0x00000001,
            DESKTOP_CREATEWINDOW = 0x00000002,
            DESKTOP_CREATEMENU = 0x00000004,
            DESKTOP_HOOKCONTROL = 0x00000008,
            DESKTOP_JOURNALRECORD = 0x00000010,
            DESKTOP_JOURNALPLAYBACK = 0x00000020,
            DESKTOP_ENUMERATE = 0x00000040,
            DESKTOP_WRITEOBJECTS = 0x00000080,
            DESKTOP_SWITCHDESKTOP = 0x00000100,

            WINSTA_ENUMDESKTOPS = 0x00000001,
            WINSTA_READATTRIBUTES = 0x00000002,
            WINSTA_ACCESSCLIPBOARD = 0x00000004,
            WINSTA_CREATEDESKTOP = 0x00000008,
            WINSTA_WRITEATTRIBUTES = 0x00000010,
            WINSTA_ACCESSGLOBALATOMS = 0x00000020,
            WINSTA_EXITWINDOWS = 0x00000040,
            WINSTA_ENUMERATE = 0x00000100,
            WINSTA_READSCREEN = 0x00000200,

            WINSTA_ALL_ACCESS = 0x0000037f
        }
        public enum SERVICE_ACCESS : uint
        {
            STANDARD_RIGHTS_REQUIRED = 0xF0000,
            SERVICE_QUERY_CONFIG = 0x00001,
            SERVICE_CHANGE_CONFIG = 0x00002,
            SERVICE_QUERY_STATUS = 0x00004,
            SERVICE_ENUMERATE_DEPENDENTS = 0x00008,
            SERVICE_START = 0x00010,
            SERVICE_STOP = 0x00020,
            SERVICE_PAUSE_CONTINUE = 0x00040,
            SERVICE_INTERROGATE = 0x00080,
            SERVICE_USER_DEFINED_CONTROL = 0x00100,
            SERVICE_ALL_ACCESS = (STANDARD_RIGHTS_REQUIRED |
                              SERVICE_QUERY_CONFIG |
                              SERVICE_CHANGE_CONFIG |
                              SERVICE_QUERY_STATUS |
                              SERVICE_ENUMERATE_DEPENDENTS |
                              SERVICE_START |
                              SERVICE_STOP |
                              SERVICE_PAUSE_CONTINUE |
                              SERVICE_INTERROGATE |
                              SERVICE_USER_DEFINED_CONTROL)
        }
        public enum SCM_ACCESS : uint
        {
            /// <summary>
            /// Required to connect to the service control manager.
            /// </summary>
            SC_MANAGER_CONNECT = 0x00001,

            /// <summary>
            /// Required to call the CreateService function to create a service
            /// object and add it to the database.
            /// </summary>
            SC_MANAGER_CREATE_SERVICE = 0x00002,

            /// <summary>
            /// Required to call the EnumServicesStatusEx function to list the
            /// services that are in the database.
            /// </summary>
            SC_MANAGER_ENUMERATE_SERVICE = 0x00004,

            /// <summary>
            /// Required to call the LockServiceDatabase function to acquire a
            /// lock on the database.
            /// </summary>
            SC_MANAGER_LOCK = 0x00008,

            /// <summary>
            /// Required to call the QueryServiceLockStatus function to retrieve
            /// the lock status information for the database.
            /// </summary>
            SC_MANAGER_QUERY_LOCK_STATUS = 0x00010,

            /// <summary>
            /// Required to call the NotifyBootConfigStatus function.
            /// </summary>
            SC_MANAGER_MODIFY_BOOT_CONFIG = 0x00020,

            /// <summary>
            /// Includes STANDARD_RIGHTS_REQUIRED, in addition to all access
            /// rights in this table.
            /// </summary>
            SC_MANAGER_ALL_ACCESS = ACCESS_MASK.STANDARD_RIGHTS_REQUIRED |
                SC_MANAGER_CONNECT |
                SC_MANAGER_CREATE_SERVICE |
                SC_MANAGER_ENUMERATE_SERVICE |
                SC_MANAGER_LOCK |
                SC_MANAGER_QUERY_LOCK_STATUS |
                SC_MANAGER_MODIFY_BOOT_CONFIG,

            GENERIC_READ = ACCESS_MASK.STANDARD_RIGHTS_READ |
                SC_MANAGER_ENUMERATE_SERVICE |
                SC_MANAGER_QUERY_LOCK_STATUS,

            GENERIC_WRITE = ACCESS_MASK.STANDARD_RIGHTS_WRITE |
                SC_MANAGER_CREATE_SERVICE |
                SC_MANAGER_MODIFY_BOOT_CONFIG,

            GENERIC_EXECUTE = ACCESS_MASK.STANDARD_RIGHTS_EXECUTE |
                SC_MANAGER_CONNECT | SC_MANAGER_LOCK,

            GENERIC_ALL = SC_MANAGER_ALL_ACCESS,
        }

        private static readonly ILog log = LogManager.GetLogger(typeof(ServiceControlManager));
        private IntPtr hServiceControlManager = IntPtr.Zero;
        private LinkedList<ENUM_SERVICE_STATUS> services = new LinkedList<ENUM_SERVICE_STATUS>();

        public bool OpenServiceControlManager()
        {
            if (this.hServiceControlManager != IntPtr.Zero)
                return true;

            IntPtr localSCM = OpenSCManager(null, null, (uint)SCM_ACCESS.SC_MANAGER_ALL_ACCESS);
            if (localSCM == IntPtr.Zero)
            {
                int errorCode = Marshal.GetLastWin32Error();
                log.Error("Error opening the ServiceControlManager: The reason is \""
                    + GetErrorMessage(errorCode)
                    + "\"");
                return false;
            }
            else
            {
                this.hServiceControlManager = localSCM;
                return true;
            }
        }

        public void EnumerateServices()
        {
            if (this.hServiceControlManager == IntPtr.Zero)
                return;

            int iBytesNeeded = 0;
            int iServicesReturned = 0;
            int iResumeHandle = 0;

            if (!EnumServicesStatus(this.hServiceControlManager,
                SERVICE_TYPES.SERVICE_WIN32,
                SERVICE_STATES.SERVICE_STATE_ALL,
                IntPtr.Zero,
                0,
                ref iBytesNeeded,
                ref iServicesReturned,
                ref iResumeHandle))
            {
                IntPtr buf = Marshal.AllocHGlobal((int)iBytesNeeded);
                if (!EnumServicesStatus(this.hServiceControlManager,
                    SERVICE_TYPES.SERVICE_WIN32,
                    SERVICE_STATES.SERVICE_STATE_ALL,
                    buf,
                    iBytesNeeded,
                    ref iBytesNeeded,
                    ref iServicesReturned,
                    ref iResumeHandle))
                {
                    int lastError = Marshal.GetLastWin32Error();
                    log.Error("EnumServicesStatus Failed "
                        + lastError
                        + ", \""
                        + GetErrorMessage(lastError)
                        + "\"");
                }
                else
                {
                    this.services.Clear();
                    ENUM_SERVICE_STATUS serviceStatus;
                    int iPtr = buf.ToInt32();
                    for (int i = 0; i < (int)iServicesReturned; i++)
                    {
                        serviceStatus = (ENUM_SERVICE_STATUS)Marshal.PtrToStructure(
                            new IntPtr(iPtr),
                            typeof(ENUM_SERVICE_STATUS));
                        this.services.AddLast(serviceStatus);
                        log.Info("Service " + serviceStatus.pServiceName);
                        iPtr += SERVICE_STATUS.SizeOf;
                    }
                    log.Info("EnumServicesStatus Success");
                }
            }
        }

        public LinkedList<ENUM_SERVICE_STATUS> getAllServices()
        {
            return this.services;
        }

        public bool StartService(String serviceName)
        {
            IntPtr hService = OpenService(serviceName);
            if (hService != IntPtr.Zero)
            {
                bool result = StartService(hService, 0, null);
                if (result == false)
                {
                    int lastError = Marshal.GetLastWin32Error();
                    String errorText = GetErrorMessage(lastError);
                    log.Error("StartService failed: ErrorCode = " + lastError +
                        ", Reason = \"" + errorText + "\"");
                }
                CloseServiceHandle(hService);
                return result;
            }
            return false;
        }

        public bool StopService(String serviceName)
        {
            IntPtr hService = OpenService(serviceName);
            if (hService != IntPtr.Zero)
            {
                SERVICE_STATUS ss = new SERVICE_STATUS();
                bool result = ControlService(hService, SERVICE_CONTROL.STOP, ref ss);
                if (result == false)
                {
                    int lastError = Marshal.GetLastWin32Error();
                    String errorText = GetErrorMessage(lastError);
                    log.Error("StopService failed: ErrorCode = " + lastError +
                        ", Reason = \"" + errorText + "\"");
                }
                CloseServiceHandle(hService);
                return result;
            }
            return false;
        }

        private IntPtr OpenService(String serviceName)
        {
            IntPtr handle = OpenService(this.hServiceControlManager, serviceName, (uint)(SERVICE_ACCESS.SERVICE_START | SERVICE_ACCESS.SERVICE_STOP | SERVICE_ACCESS.STANDARD_RIGHTS_REQUIRED));

            if (handle == IntPtr.Zero)
            {
                int lastError = Marshal.GetLastWin32Error();
                String errorText = GetErrorMessage(lastError);
                log.Error( "OpenService failed: ErrorCode = " + lastError +
                    ", Reason = \"" + errorText + "\""); 
            }
            return handle;
        }

        public void Dispose()
        {
            if (this.hServiceControlManager != IntPtr.Zero)
            {
                CloseServiceHandle(this.hServiceControlManager);
            }
        }

        public static string GetErrorMessage(int errorCode)
        {
            int FORMAT_MESSAGE_ALLOCATE_BUFFER = 0x00000100;
            int FORMAT_MESSAGE_IGNORE_INSERTS = 0x00000200;
            int FORMAT_MESSAGE_FROM_SYSTEM = 0x00001000;

            int messageSize = 255;
            string lpMsgBuf = "";
            int dwFlags = FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS;

            int retVal = FormatMessage(dwFlags, 0, errorCode, 0, ref lpMsgBuf, messageSize, 0);
            if (0 == retVal)
            {
                return null;
            }
            else
            {
                return lpMsgBuf;
            }
        }

    }
}
